from __future__ import annotations

from ipaddress import IPv4Address

from tests.conftest import SKIPIF_CI_AND_MAC
from utilities.dnspython import nslookup


class TestNSLookup:
    @SKIPIF_CI_AND_MAC
    def test_main(self) -> None:
        result = nslookup("localhost")
        expected = [IPv4Address("127.0.0.1")]
        assert result == expected



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from tests.conftest import SKIPIF_CI
from utilities.json import run_prettier, write_formatted_json

if TYPE_CHECKING:
    from pathlib import Path


class TestRunPrettier:
    input_: ClassVar[str] = '{"foo":0,"bar":[1,2,3]}'
    output: ClassVar[str] = '{ "foo": 0, "bar": [1, 2, 3] }\n'

    @SKIPIF_CI
    def test_bytes(self) -> None:
        result = run_prettier(self.input_.encode())
        assert result == self.output.encode()

    @SKIPIF_CI
    def test_text(self) -> None:
        result = run_prettier(self.input_)
        assert result == self.output

    @SKIPIF_CI
    def test_path(self, *, tmp_path: Path) -> None:
        file = tmp_path.joinpath("file.json")
        _ = file.write_text(self.input_)
        run_prettier(file)
        contents = file.read_text()
        assert contents == self.output

    @SKIPIF_CI
    def test_file_as_str(self, *, tmp_path: Path) -> None:
        file = tmp_path.joinpath("file.json")
        _ = file.write_text(self.input_)
        _ = run_prettier(str(file))
        contents = file.read_text()
        assert contents == self.output


class TestWriteFormattedJSON:
    input_: ClassVar[str] = '{"foo":0,"bar":[1,2,3]}'
    output: ClassVar[str] = '{ "foo": 0, "bar": [1, 2, 3] }\n'

    @SKIPIF_CI
    def test_main(self, *, tmp_path: Path) -> None:
        file = tmp_path.joinpath("file.json")
        write_formatted_json(self.input_.encode(), file)
        contents = file.read_text()
        assert contents == self.output



from __future__ import annotations

from functools import cache as _cache
from functools import lru_cache as _lru_cache
from functools import partial as _partial
from typing import TYPE_CHECKING, Any, cast, overload, override

if TYPE_CHECKING:
    from collections.abc import Callable


def cache[F: Callable](func: F, /) -> F:
    """Typed version of `cache`."""
    typed_cache = cast("Callable[[F], F]", _cache)
    return typed_cache(func)


##


_MAX_SIZE = 128


@overload
def lru_cache[F: Callable](
    func: F, /, *, max_size: int = _MAX_SIZE, typed: bool = False
) -> F: ...
@overload
def lru_cache[F: Callable](
    func: None = None, /, *, max_size: int = _MAX_SIZE, typed: bool = False
) -> Callable[[F], F]: ...
def lru_cache[F: Callable](
    func: F | None = None, /, *, max_size: int = _MAX_SIZE, typed: bool = False
) -> F | Callable[[F], F]:
    """Typed version of `lru_cache`."""
    if func is None:
        result = partial(lru_cache, max_size=max_size, typed=typed)
        return cast("Callable[[F], F]", result)
    wrapped = _lru_cache(maxsize=max_size, typed=typed)(func)
    return cast("Any", wrapped)


##


class partial[T](_partial[T]):  # noqa: N801
    """Partial which accepts Ellipsis for positional arguments."""

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> T:
        iter_args = iter(args)
        head = (next(iter_args) if arg is ... else arg for arg in self.args)
        return self.func(*head, *iter_args, **{**self.keywords, **kwargs})


__all__ = ["cache", "lru_cache", "partial"]



from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING
from zipfile import ZipFile

from utilities.tempfile import TemporaryDirectory

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from utilities.types import PathLike


@contextmanager
def yield_zip_file_contents(path: PathLike, /) -> Iterator[list[Path]]:
    """Yield the contents of a zipfile in a temporary directory."""
    with ZipFile(path) as zf, TemporaryDirectory() as temp:
        zf.extractall(path=temp)
        yield list(temp.iterdir())
    _ = zf  # make coverage understand this is returned


__all__ = ["yield_zip_file_contents"]



from __future__ import annotations

from asyncio import Queue, TaskGroup, run
from collections.abc import ItemsView, KeysView, ValuesView
from contextlib import asynccontextmanager
from re import search
from typing import TYPE_CHECKING, ClassVar

from hypothesis import given
from hypothesis.strategies import booleans, dictionaries, integers, lists, none
from pytest import LogCaptureFixture, RaisesGroup, mark, param, raises

from utilities.asyncio import (
    AsyncDict,
    EnhancedTaskGroup,
    get_coroutine_name,
    get_items,
    get_items_nowait,
    loop_until_succeed,
    put_items,
    put_items_nowait,
    sleep_max,
    sleep_rounded,
    sleep_td,
    sleep_until,
    stream_command,
    timeout_td,
)
from utilities.hypothesis import pairs, text_ascii
from utilities.pytest import skipif_windows
from utilities.text import unique_str
from utilities.timer import Timer
from utilities.whenever import MILLISECOND, SECOND, get_now

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from whenever import TimeDelta


async_dicts = dictionaries(text_ascii(), integers()).map(AsyncDict)


class TestAsyncDict:
    @given(dict_=async_dicts)
    async def test_aenter(self, *, dict_: AsyncDict[str, int]) -> None:
        async with dict_:
            ...

    @given(dict_=async_dicts)
    async def test_clear(self, *, dict_: AsyncDict[str, int]) -> None:
        await dict_.clear()
        assert len(dict_) == 0

    @given(dict_=async_dicts, key=text_ascii())
    def test_contains(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        assert isinstance(key in dict_, bool)

    @given(dict_=async_dicts)
    def test_copy(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.copy(), AsyncDict)

    @given(dict_=async_dicts, key=text_ascii())
    async def test_del(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        if key in dict_:
            await dict_.del_(key)
        else:
            with raises(KeyError):
                await dict_.del_(key)

    @given(dict_=async_dicts)
    def test_empty(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.empty, bool)

    @given(dicts=pairs(async_dicts))
    def test_eq(
        self, *, dicts: tuple[AsyncDict[str, int], AsyncDict[str, int]]
    ) -> None:
        first, second = dicts
        assert isinstance(first == second, bool)

    @given(keys=lists(text_ascii()))
    def test_fromkeys(self, *, keys: list[str]) -> None:
        dict_ = AsyncDict.fromkeys(keys)
        assert isinstance(dict_, AsyncDict)

    @given(dict_=async_dicts, key=text_ascii())
    def test_get(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        if key in dict_:
            assert isinstance(dict_.get(key), int)
        else:
            assert dict_.get(key) is None

    @given(dict_=async_dicts, key=text_ascii())
    def test_get_default(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        value = dict_.get(key, None)
        assert isinstance(value, int) or (value is None)

    @given(dict_=async_dicts, key=text_ascii())
    def test_getitem(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        if key in dict_:
            assert isinstance(dict_[key], int)
        else:
            with raises(KeyError):
                _ = dict_[key]

    @given(dict_=async_dicts)
    def test_items(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.items(), ItemsView)
        for key, value in dict_.items():
            assert isinstance(key, str)
            assert isinstance(value, int)

    @given(dict_=async_dicts)
    def test_iter(self, *, dict_: AsyncDict[str, int]) -> None:
        for key in dict_:
            assert isinstance(key, str)

    @given(dict_=async_dicts)
    def test_keys(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.keys(), KeysView)
        for key in dict_.keys():  # noqa: SIM118
            assert isinstance(key, str)

    @given(dict_=async_dicts)
    def test_len(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(len(dict_), int)

    @given(dict_=async_dicts, key=text_ascii())
    async def test_pop(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        if key in dict_:
            assert isinstance(await dict_.pop(key), int)
        else:
            with raises(KeyError):
                _ = await dict_.pop(key)

    @given(dict_=async_dicts, key=text_ascii())
    async def test_pop_default(self, *, dict_: AsyncDict[str, int], key: str) -> None:
        value = await dict_.pop(key, None)
        assert isinstance(value, int) or (value is None)

    @given(dict_=async_dicts)
    async def test_popitem(self, *, dict_: AsyncDict[str, int]) -> None:
        if len(dict_) >= 1:
            key, value = await dict_.popitem()
            assert isinstance(key, str)
            assert isinstance(value, int)
        else:
            with raises(KeyError):
                _ = await dict_.popitem()

    @given(dict_=async_dicts)
    def test_repr(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(repr(dict_), str)

    @given(dict_=async_dicts)
    def test_reversed(self, *, dict_: AsyncDict[str, int]) -> None:
        for key in reversed(dict_):
            assert isinstance(key, str)

    @given(dict_=async_dicts, key=text_ascii(), value=integers())
    async def test_set(
        self, *, dict_: AsyncDict[str, int], key: str, value: int
    ) -> None:
        await dict_.set(key, value)

    @given(dict_=async_dicts, key=text_ascii(), value=integers())
    async def test_setdefault(
        self, *, dict_: AsyncDict[str, int], key: str, value: int
    ) -> None:
        assert isinstance(await dict_.setdefault(key, value), int)

    @given(dict_=async_dicts)
    def test_str(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(str(dict_), str)

    @given(dicts=pairs(async_dicts))
    async def test_update(
        self, *, dicts: tuple[AsyncDict[str, int], AsyncDict[str, int]]
    ) -> None:
        first, second = dicts
        await first.update(second)

    @given(dict_=async_dicts)
    def test_values(self, *, dict_: AsyncDict[str, int]) -> None:
        assert isinstance(dict_.values(), ValuesView)
        for value in dict_.values():
            assert isinstance(value, int)


class TestEnhancedTaskGroup:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    async def test_create_task_context_coroutine(self) -> None:
        flag: bool = False

        @asynccontextmanager
        async def yield_true() -> AsyncIterator[None]:
            nonlocal flag
            try:
                flag = True
                yield
            finally:
                flag = False

        assert not flag
        async with EnhancedTaskGroup(timeout=2 * self.delta) as tg:
            _ = tg.create_task_context(yield_true())
            await sleep_td(self.delta)
            assert flag
        assert not flag

    async def test_max_tasks_disabled(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup() as tg:
                for _ in range(10):
                    _ = tg.create_task(sleep_td(self.delta))
        assert timer <= 2 * self.delta

    async def test_max_tasks_enabled(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=2) as tg:
                for _ in range(10):
                    _ = tg.create_task(sleep_td(self.delta))
        assert timer >= 5 * self.delta

    async def test_run_or_create_many_tasks_parallel_with_max_tasks_two(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=2) as tg:
                assert not tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_many_tasks(sleep_td, self.delta)
        assert timer >= 5 * self.delta

    async def test_run_or_create_many_tasks_serial_with_debug(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(debug=True) as tg:
                assert tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_many_tasks(sleep_td, self.delta)
        assert timer >= 10 * self.delta

    async def test_run_or_create_task_parallel_with_max_tasks_none(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup() as tg:
                assert not tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_task(sleep_td(self.delta))
        assert timer <= 2 * self.delta

    async def test_run_or_create_task_parallel_with_max_tasks_two(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=2) as tg:
                assert not tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_task(sleep_td(self.delta))
        assert timer >= 5 * self.delta

    async def test_run_or_create_task_serial_with_max_tasks_negative(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=-1) as tg:
                assert tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_task(sleep_td(self.delta))
        assert timer >= 10 * self.delta

    async def test_run_or_create_task_serial_with_debug(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(debug=True) as tg:
                assert tg._is_debug()
                for _ in range(10):
                    _ = await tg.run_or_create_task(sleep_td(self.delta))
        assert timer >= 10 * self.delta

    async def test_timeout_pass(self) -> None:
        async with EnhancedTaskGroup(timeout=2 * self.delta) as tg:
            _ = tg.create_task(sleep_td(self.delta))

    async def test_timeout_fail(self) -> None:
        with RaisesGroup(TimeoutError):
            async with EnhancedTaskGroup(timeout=self.delta) as tg:
                _ = tg.create_task(sleep_td(2 * self.delta))

    async def test_custom_error(self) -> None:
        class CustomError(Exception): ...

        with RaisesGroup(CustomError):
            async with EnhancedTaskGroup(timeout=self.delta, error=CustomError) as tg:
                _ = tg.create_task(sleep_td(2 * self.delta))


class TestGetCoroutineName:
    def test_main(self) -> None:
        async def func() -> None:
            return None

        result = get_coroutine_name(func)
        expected = "func"
        assert result == expected


class TestGetItems:
    @given(
        xs=lists(integers(), min_size=1),
        max_size=integers(1, 10) | none(),
        wait=booleans(),
    )
    async def test_main(
        self, *, xs: list[int], max_size: int | None, wait: bool
    ) -> None:
        queue: Queue[int] = Queue()
        put_items_nowait(xs, queue)
        if wait:
            result = await get_items(queue, max_size=max_size)
        else:
            result = get_items_nowait(queue, max_size=max_size)
        assert result == xs[:max_size]


class TestLoopUntilSucceed:
    @mark.parametrize("sleep", [param(MILLISECOND), param(None)])
    @mark.parametrize("use_logger", [param(True), param(False)])
    async def test_main(
        self, *, caplog: LogCaptureFixture, sleep: TimeDelta | None, use_logger: bool
    ) -> None:
        class CustomError(Exception): ...

        caplog.set_level("DEBUG", logger=(name := unique_str()))
        counter = 0

        async def func() -> None:
            nonlocal counter
            counter += 1
            if counter <= 3:
                raise CustomError

        assert await loop_until_succeed(
            lambda: func(), logger=name if use_logger else None, sleep=sleep
        )
        assert counter == 4

        if use_logger:
            messages = [r.message for r in caplog.records if r.name == name]
            expected = 3 * (
                ["Error running 'func'"]
                + ([] if sleep is None else ["Sleeping for PT0.001S..."])
                + ["Retrying 'func'..."]
            )
            assert messages == expected

    async def test_error_flat(self) -> None:
        class CustomError(Exception): ...

        counter = 0

        async def func() -> None:
            nonlocal counter
            counter += 1
            if counter <= 3:
                raise CustomError

        assert not await loop_until_succeed(lambda: func(), errors=CustomError)
        assert counter == 1

    async def test_error_nested(self) -> None:
        class CustomError(Exception): ...

        counter = 0

        async def func() -> None:
            async with TaskGroup() as tg:
                _ = tg.create_task(inner())

        async def inner() -> None:
            nonlocal counter
            counter += 1
            if counter <= 3:
                raise CustomError

        assert not await loop_until_succeed(lambda: func(), errors=CustomError)
        assert counter == 1


class TestPutItems:
    @given(xs=lists(integers(), min_size=1), wait=booleans())
    async def test_main(self, *, xs: list[int], wait: bool) -> None:
        queue: Queue[int] = Queue()
        if wait:
            put_items_nowait(xs, queue)
        else:
            await put_items(xs, queue)
        result: list[int] = []
        while not queue.empty():
            result.append(await queue.get())
        assert result == xs


class TestSleepMaxDur:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    async def test_main(self) -> None:
        with Timer() as timer:
            await sleep_max(self.delta)
        assert timer <= 2 * self.delta

    async def test_none(self) -> None:
        with Timer() as timer:
            await sleep_max()
        assert timer <= self.delta


class TestSleepTD:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    async def test_main(self) -> None:
        with Timer() as timer:
            await sleep_td(self.delta)
        assert timer <= 2 * self.delta

    async def test_none(self) -> None:
        with Timer() as timer:
            await sleep_td()
        assert timer <= self.delta


class TestSleepUntil:
    async def test_main(self) -> None:
        await sleep_until(get_now() + 0.05 * SECOND)


class TestSleepUntilRounded:
    async def test_main(self) -> None:
        await sleep_rounded(10 * MILLISECOND)


class TestStreamCommand:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    @skipif_windows
    async def test_main(self) -> None:
        output = await stream_command(
            'echo "stdout message" && sleep 0.1 && echo "stderr message" >&2'
        )
        await sleep_td(self.delta)
        assert output.return_code == 0
        assert output.stdout == "stdout message\n"
        assert output.stderr == "stderr message\n"

    @skipif_windows
    async def test_error(self) -> None:
        output = await stream_command("this-is-an-error")
        await sleep_td(self.delta)
        assert output.return_code == 127
        assert output.stdout == ""
        assert search(
            r"^/bin/sh: (1: )?this-is-an-error: (command )?not found$", output.stderr
        )


class TestTimeoutTD:
    delta: ClassVar[TimeDelta] = 0.05 * SECOND

    async def test_pass(self) -> None:
        async with timeout_td(2 * self.delta):
            await sleep_td(self.delta)

    async def test_fail(self) -> None:
        with raises(TimeoutError):
            async with timeout_td(self.delta):
                await sleep_td(2 * self.delta)

    async def test_custom_error(self) -> None:
        class CustomError(Exception): ...

        with raises(CustomError):
            async with timeout_td(self.delta, error=CustomError):
                await sleep_td(2 * self.delta)


if __name__ == "__main__":
    _ = run(
        stream_command('echo "stdout message" && sleep 2 && echo "stderr message" >&2')
    )



from __future__ import annotations

import datetime as dt
from math import inf
from typing import TYPE_CHECKING

import polars as pl
from hypothesis import HealthCheck, given, settings
from hypothesis.strategies import (
    booleans,
    floats,
    integers,
    just,
    lists,
    none,
    sampled_from,
    tuples,
)
from polars import DataFrame, Float64, datetime_range, int_range
from pytest import fixture

from utilities.altair import (
    plot_dataframes,
    plot_intraday_dataframe,
    save_chart,
    save_charts_as_pdf,
    vconcat_charts,
)
from utilities.functions import ensure_class
from utilities.hypothesis import dates, text_ascii, zoned_datetimes
from utilities.polars import DatetimeUTC, zoned_datetime
from utilities.tzdata import HongKong, Tokyo
from utilities.whenever import get_now
from utilities.zoneinfo import UTC

if TYPE_CHECKING:
    from pathlib import Path
    from zoneinfo import ZoneInfo

    import whenever
    from whenever import ZonedDateTime


@fixture
def time_series() -> DataFrame:
    return (
        datetime_range(
            dt.datetime(2024, 1, 1, tzinfo=UTC),
            dt.datetime(2024, 1, 7, 23, tzinfo=UTC),
            interval="1h",
            eager=True,
        )
        .rename("datetime")
        .to_frame()
        .with_columns(x=int_range(end=pl.len()), y=int_range(end=2 * pl.len(), step=2))
    )


class TestPlotDataFrames:
    @given(
        x=just("datetime") | none(),
        y=sampled_from([
            "x",
            "y",
            ("x", 50),
            ("y", 50),
            (["x", "y"], 50),
            ["x", "y"],
            [["x"], ["y"]],
            [["x"], ("y", 50)],
        ])
        | none(),
        height=integers(1, 100),
        width=integers(1, 100),
        title=text_ascii() | none(),
    )
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    def test_main(
        self,
        *,
        time_series: DataFrame,
        x: str | None,
        y: str
        | tuple[str, int]
        | tuple[list[str], int]
        | list[str | list[str] | tuple[str, int] | tuple[list[str], int]]
        | None,
        height: int,
        width: int,
        title: str | None,
    ) -> None:
        _ = plot_dataframes(
            time_series, x=x, y=y, height=height, width=width, title=title
        )

    @given(time_zone=sampled_from([UTC, HongKong, Tokyo]))
    def test_auto_localization(self, *, time_zone: ZoneInfo) -> None:
        df = DataFrame(
            data=[(dt.datetime(2000, 1, 1, 12, tzinfo=time_zone), 0.0)],
            schema={"datetime": zoned_datetime(time_zone=time_zone), "value": Float64},
            orient="row",
        )
        chart = plot_dataframes(df, x="datetime", y="value")
        datetime = ensure_class(chart.data, DataFrame)["datetime"].item()
        expected = dt.datetime(2000, 1, 1, 12, tzinfo=time_zone).replace(tzinfo=None)
        assert datetime == expected

    @given(data=lists(tuples(dates(), floats(-10, 10))))
    def test_tooltip_format_date(
        self, *, data: list[tuple[whenever.Date, float]]
    ) -> None:
        df = DataFrame(
            data=[(d.py_date(), v) for d, v in data],
            schema={"index": pl.Date, "value": Float64},
            orient="row",
        )
        _ = plot_dataframes(df, x="index", y="value")

    @given(data=lists(tuples(zoned_datetimes(), floats(-10, 10))))
    def test_tooltip_format_date_time(
        self, *, data: list[tuple[ZonedDateTime, float]]
    ) -> None:
        df = DataFrame(
            data=[(d.py_datetime(), v) for d, v in data],
            schema={"index": DatetimeUTC, "value": Float64},
            orient="row",
        )
        _ = plot_dataframes(df, x="index", y="value")


class TestPlotIntradayDataFrame:
    @given(interactive=booleans(), width=integers(1, 100))
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    def test_main(
        self, *, time_series: DataFrame, interactive: bool, width: int
    ) -> None:
        _ = plot_intraday_dataframe(time_series, interactive=interactive, width=width)

    def test_non_finite(self) -> None:
        data = DataFrame(
            data=[(get_now().py_datetime(), inf)],
            schema={"datetime": DatetimeUTC, "value": Float64},
            orient="row",
        )
        _ = plot_intraday_dataframe(data)


class TestSaveChart:
    def test_main(self, *, time_series: DataFrame, tmp_path: Path) -> None:
        chart = plot_dataframes(time_series)
        save_chart(chart, tmp_path.joinpath("chart.png"))


class TestSaveChartsAsPdf:
    def test_main(self, *, time_series: DataFrame, tmp_path: Path) -> None:
        chart = plot_dataframes(time_series)
        save_charts_as_pdf(chart, path=tmp_path.joinpath("chart.pdf"))


class TestVConcatCharts:
    @given(width=integers(1, 100))
    @settings(suppress_health_check={HealthCheck.function_scoped_fixture})
    def test_main(self, *, time_series: DataFrame, width: int) -> None:
        chart1 = plot_intraday_dataframe(time_series, interactive=False)
        chart2 = plot_intraday_dataframe(time_series, interactive=False)
        _ = vconcat_charts(chart1, chart2, width=width)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, assert_never, override

if TYPE_CHECKING:
    from utilities.types import ExceptionTypeLike, MaybeType


@dataclass(kw_only=True, slots=True)
class ImpossibleCaseError(Exception):
    case: list[str]

    @override
    def __str__(self) -> str:
        desc = ", ".join(self.case)
        return f"Case must be possible: {desc}."
        ##


##


def is_instance_error(
    error: BaseException, class_or_tuple: ExceptionTypeLike[Exception], /
) -> bool:
    """Check if an instance relationship holds, allowing for groups."""
    if isinstance(error, class_or_tuple):
        return True
    if not isinstance(error, BaseExceptionGroup):
        return False
    return any(is_instance_error(e, class_or_tuple) for e in error.exceptions)


##


def repr_error(error: MaybeType[BaseException], /) -> str:
    """Get a string representation of an error."""
    match error:
        case ExceptionGroup() as group:
            descs = list(map(repr_error, group.exceptions))
            joined = ", ".join(descs)
            return f"{group.__class__.__name__}({joined})"
        case BaseException() as error_obj:
            return f"{error_obj.__class__.__name__}({error_obj})"
        case type() as error_cls:
            return error_cls.__name__
        case _ as never:
            assert_never(never)


__all__ = ["ImpossibleCaseError", "is_instance_error", "repr_error"]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from random import Random


try:
    from pytest import fixture
except ModuleNotFoundError:
    pass
else:

    @fixture
    def random_state(*, seed: int) -> Random:
        """Fixture for a random state."""
        from utilities.random import get_state

        return get_state(seed=seed)


__all__ = ["random_state"]



from __future__ import annotations

from re import search
from typing import TYPE_CHECKING
from uuid import UUID

from hypothesis import given
from hypothesis.strategies import integers, none, randoms, uuids

from utilities.uuid import UUID_EXACT_PATTERN, UUID_PATTERN, get_uuid

if TYPE_CHECKING:
    from random import Random


class TestGetUUID4:
    @given(seed=randoms() | none())
    def test_main(self, *, seed: Random | None) -> None:
        uuid = get_uuid(seed=seed)
        assert isinstance(uuid, UUID)

    @given(seed=integers())
    def test_deterministic(self, *, seed: int) -> None:
        uuid1, uuid2 = [get_uuid(seed=seed) for _ in range(2)]
        assert uuid1 == uuid2


class TestUUIDPattern:
    @given(uuid=uuids())
    def test_main(self, *, uuid: UUID) -> None:
        assert search(UUID_PATTERN, str(uuid))

    @given(uuid=uuids())
    def test_exact(self, *, uuid: UUID) -> None:
        text = f".{uuid}."
        assert search(UUID_PATTERN, text)
        assert not search(UUID_EXACT_PATTERN, text)



from __future__ import annotations

import datetime as dt
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from types import NoneType, UnionType
from typing import (
    Any,
    Literal,
    NamedTuple,
    Optional,  # pyright: ignore[reportDeprecated]
    TypeAliasType,
    TypeGuard,
    Union,  # pyright: ignore[reportDeprecated]
    get_origin,
    overload,
    override,
)
from typing import get_args as _get_args
from typing import get_type_hints as _get_type_hints
from uuid import UUID
from warnings import warn

from whenever import (
    Date,
    DateDelta,
    DateTimeDelta,
    PlainDateTime,
    Time,
    TimeDelta,
    ZonedDateTime,
)

from utilities.iterables import unique_everseen
from utilities.sentinel import Sentinel
from utilities.types import StrMapping


def get_args(obj: Any, /, *, optional_drop_none: bool = False) -> tuple[Any, ...]:
    """Get the arguments of an annotation."""
    if isinstance(obj, TypeAliasType):
        return get_args(obj.__value__)
    args = _get_args(obj)
    if is_optional_type(obj) and optional_drop_none:
        args = tuple(a for a in args if a is not NoneType)
    return args


##


def get_literal_elements(obj: Any, /) -> list[Any]:
    """Get the elements of a literal annotation."""
    return _get_literal_elements_inner(obj)


def _get_literal_elements_inner(obj: Any, /) -> list[Any]:
    if isinstance(obj, str | int):
        return [obj]
    args = get_args(obj)
    parts = chain.from_iterable(map(_get_literal_elements_inner, args))
    return list(unique_everseen(parts))


##


def get_type_classes(obj: Any, /) -> tuple[type[Any], ...]:
    """Get the type classes from a type/tuple/Union type."""
    types: Sequence[type[Any]] = []
    if isinstance(obj, type):
        types.append(obj)
    elif isinstance(obj, tuple):
        for arg in obj:
            if isinstance(arg, type):
                types.append(arg)
            elif isinstance(arg, tuple):
                types.extend(get_type_classes(arg))
            elif is_union_type(arg):
                types.extend(get_union_type_classes(arg))
            else:
                raise _GetTypeClassesTupleError(obj=obj, inner=arg)
    elif is_union_type(obj):
        types.extend(get_union_type_classes(obj))
    else:
        raise _GetTypeClassesTypeError(obj=obj)
    return tuple(types)


@dataclass(kw_only=True, slots=True)
class GetTypeClassesError(Exception):
    obj: Any


@dataclass(kw_only=True, slots=True)
class _GetTypeClassesTypeError(GetTypeClassesError):
    @override
    def __str__(self) -> str:
        return f"Object must be a type, tuple or Union type; got {self.obj!r} of type {type(self.obj)!r}"


@dataclass(kw_only=True, slots=True)
class _GetTypeClassesTupleError(GetTypeClassesError):
    inner: Any

    @override
    def __str__(self) -> str:
        return f"Tuple must contain types, tuples or Union types only; got {self.inner} of type {type(self.inner)!r}"


##


def get_type_hints(
    obj: Any,
    /,
    *,
    globalns: StrMapping | None = None,
    localns: StrMapping | None = None,
    warn_name_errors: bool = False,
) -> dict[str, Any]:
    """Get the type hints of an object."""
    result: dict[str, Any] = obj.__annotations__
    _ = {
        Date,
        DateDelta,
        DateTimeDelta,
        Literal,
        Path,
        PlainDateTime,
        Sentinel,
        StrMapping,
        Time,
        TimeDelta,
        UUID,
        ZonedDateTime,
        dt,
    }
    globalns_use = globals() | ({} if globalns is None else dict(globalns))
    localns_use = {} if localns is None else dict(localns)
    try:
        hints = _get_type_hints(obj, globalns=globalns_use, localns=localns_use)
    except NameError as error:
        if warn_name_errors:
            warn(f"Error getting type hints for {obj!r}; {error}", stacklevel=2)
    else:
        result.update({
            key: value
            for key, value in hints.items()
            if (key not in result) or isinstance(result[key], str)
        })
    return result


##


def get_union_type_classes(obj: Any, /) -> tuple[type[Any], ...]:
    """Get the type classes from a Union type."""
    if not is_union_type(obj):
        raise _GetUnionTypeClassesUnionTypeError(obj=obj)
    types_: Sequence[type[Any]] = []
    for arg in get_args(obj):
        if not isinstance(arg, type):
            raise _GetUnionTypeClassesInternalTypeError(obj=obj, inner=arg)
        types_.append(arg)
    return tuple(types_)


@dataclass(kw_only=True, slots=True)
class GetUnionTypeClassesError(Exception):
    obj: Any


@dataclass(kw_only=True, slots=True)
class _GetUnionTypeClassesUnionTypeError(GetUnionTypeClassesError):
    @override
    def __str__(self) -> str:
        return (
            f"Object must be a Union type; got {self.obj!r} of type {type(self.obj)!r}"
        )


@dataclass(kw_only=True, slots=True)
class _GetUnionTypeClassesInternalTypeError(GetUnionTypeClassesError):
    inner: Any

    @override
    def __str__(self) -> str:
        return f"Union type must contain types only; got {self.inner} of type {type(self.inner)!r}"


##


def is_dict_type(obj: Any, /) -> bool:
    """Check if an object is a dict type annotation."""
    return _is_annotation_of_type(obj, dict)


##


def is_frozenset_type(obj: Any, /) -> bool:
    """Check if an object is a frozenset type annotation."""
    return _is_annotation_of_type(obj, frozenset)


##


@overload
def is_instance_gen[T](obj: Any, type_: type[T], /) -> TypeGuard[T]: ...
@overload
def is_instance_gen[T1](obj: Any, type_: tuple[T1], /) -> TypeGuard[T1]: ...
@overload
def is_instance_gen[T1, T2](
    obj: Any, type_: tuple[T1, T2], /
) -> TypeGuard[T1 | T2]: ...
@overload
def is_instance_gen[T1, T2, T3](
    obj: Any, type_: tuple[T1, T2, T3], /
) -> TypeGuard[T1 | T2 | T3]: ...
@overload
def is_instance_gen[T1, T2, T3, T4](
    obj: Any, type_: tuple[T1, T2, T3, T4], /
) -> TypeGuard[T1 | T2 | T3 | T4]: ...
@overload
def is_instance_gen[T1, T2, T3, T4, T5](
    obj: Any, type_: tuple[T1, T2, T3, T4, T5], /
) -> TypeGuard[T1 | T2 | T3 | T4 | T5]: ...
@overload
def is_instance_gen(obj: Any, type_: Any, /) -> bool: ...
def is_instance_gen(obj: Any, type_: Any, /) -> bool:
    """Check if an instance relationship holds, except bool<int."""
    # parent
    if isinstance(type_, tuple):
        return any(is_instance_gen(obj, t) for t in type_)  # skipif-ci-and-not-windows
    if is_literal_type(type_):
        return obj in get_args(type_)
    if is_union_type(type_):
        return any(is_instance_gen(obj, t) for t in get_args(type_))
    # tuple vs tuple
    if isinstance(obj, tuple) and is_tuple_type(type_):
        type_args = get_args(type_)
        return (len(obj) == len(type_args)) and all(
            is_instance_gen(o, t) for o, t in zip(obj, type_args, strict=True)
        )
    if isinstance(obj, tuple) is not is_tuple_type(type_):
        return False
    # basic
    if isinstance(type_, type):
        return any(_is_instance_gen_type(obj, t) for t in get_type_classes(type_))
    raise IsInstanceGenError(obj=obj, type_=type_)


def _is_instance_gen_type[T](obj: Any, type_: type[T], /) -> TypeGuard[T]:
    return (
        isinstance(obj, type_)
        and not (
            isinstance(obj, bool)
            and issubclass(type_, int)
            and not issubclass(type_, bool)
        )
        and not (
            isinstance(obj, dt.datetime)
            and issubclass(type_, dt.date)
            and not issubclass(type_, dt.datetime)
        )
    )


@dataclass(kw_only=True, slots=True)
class IsInstanceGenError(Exception):
    obj: Any
    type_: Any

    @override
    def __str__(self) -> str:
        return f"Invalid arguments; got {self.obj!r} of type {type(self.obj)!r} and {self.type_!r} of type {type(self.type_)!r}"


##


def is_list_type(obj: Any, /) -> bool:
    """Check if an object is a list type annotation."""
    return _is_annotation_of_type(obj, list)


##


def is_literal_type(obj: Any, /) -> bool:
    """Check if an object is a literal type annotation."""
    return _is_annotation_of_type(obj, Literal)


##


def is_mapping_type(obj: Any, /) -> bool:
    """Check if an object is a mapping type annotation."""
    return _is_annotation_of_type(obj, Mapping)


##


def is_namedtuple_class(obj: Any, /) -> TypeGuard[type[Any]]:
    """Check if an object is a namedtuple."""
    return isinstance(obj, type) and _is_namedtuple_core(obj)


def is_namedtuple_instance(obj: Any, /) -> bool:
    """Check if an object is an instance of a dataclass."""
    return (not isinstance(obj, type)) and _is_namedtuple_core(obj)


def _is_namedtuple_core(obj: Any, /) -> bool:
    """Check if an object is an instance of a dataclass."""
    try:
        (base,) = obj.__orig_bases__
    except (AttributeError, ValueError):
        return False
    return base is NamedTuple


##


def is_optional_type(obj: Any, /) -> bool:
    """Check if an object is an optional type annotation."""
    is_optional = _is_annotation_of_type(obj, Optional)  # pyright: ignore[reportDeprecated]
    return is_optional or (
        is_union_type(obj) and any(a is NoneType for a in _get_args(obj))
    )


##


def is_sequence_type(obj: Any, /) -> bool:
    """Check if an object is a sequence type annotation."""
    return _is_annotation_of_type(obj, Sequence)


##


def is_set_type(obj: Any, /) -> bool:
    """Check if an object is a set type annotation."""
    return _is_annotation_of_type(obj, set)


##


@overload
def is_subclass_gen[T](cls: type[Any], parent: type[T], /) -> TypeGuard[type[T]]: ...
@overload
def is_subclass_gen[T1](
    cls: type[Any], parent: tuple[type[T1]], /
) -> TypeGuard[type[T1]]: ...
@overload
def is_subclass_gen[T1, T2](
    cls: type[Any], parent: tuple[type[T1], type[T2]], /
) -> TypeGuard[type[T1 | T2]]: ...
@overload
def is_subclass_gen[T1, T2, T3](
    cls: type[Any], parent: tuple[type[T1], type[T2], type[T3]], /
) -> TypeGuard[type[T1 | T2 | T3]]: ...
@overload
def is_subclass_gen[T1, T2, T3, T4](
    cls: type[Any], parent: tuple[type[T1], type[T2], type[T3], type[T4]], /
) -> TypeGuard[type[T1 | T2 | T3 | T4]]: ...
@overload
def is_subclass_gen[T1, T2, T3, T4, T5](
    cls: type[Any], parent: tuple[type[T1], type[T2], type[T3], type[T4], type[T5]], /
) -> TypeGuard[type[T1 | T2 | T3 | T4 | T5]]: ...
@overload
def is_subclass_gen(cls: Any, parent: Any, /) -> bool: ...
def is_subclass_gen(cls: Any, parent: Any, /) -> bool:
    """Generalized `issubclass`."""
    # child
    if isinstance(cls, tuple):
        return all(is_subclass_gen(c, parent) for c in cls)
    if is_literal_type(cls):
        types = tuple(map(type, get_args(cls)))
        return (
            is_literal_type(parent) and set(get_args(cls)).issubset(get_args(parent))
        ) or is_subclass_gen(types, parent)
    if is_union_type(cls):
        return all(is_subclass_gen(c, parent) for c in get_args(cls))
    # parent
    if isinstance(parent, tuple):
        return any(is_subclass_gen(cls, p) for p in parent)
    if is_literal_type(parent):
        return is_literal_type(cls) and set(get_args(cls)).issubset(get_args(parent))
    if is_union_type(parent):
        return any(is_subclass_gen(cls, p) for p in get_args(parent))
    # tuple vs tuple
    if is_tuple_type(cls) and is_tuple_type(parent):
        cls_args, parent_args = get_args(cls), get_args(parent)
        return (len(cls_args) == len(parent_args)) and all(
            is_subclass_gen(c, p) for c, p in zip(cls_args, parent_args, strict=True)
        )
    if is_tuple_type(cls) is not is_tuple_type(parent):
        return False
    # basic
    if isinstance(cls, type):
        return any(_is_subclass_gen_type(cls, p) for p in get_type_classes(parent))
    raise IsSubclassGenError(cls=cls)


def _is_subclass_gen_type[T](cls: type[Any], parent: type[T], /) -> TypeGuard[type[T]]:
    return (
        issubclass(cls, parent)
        and not (
            issubclass(cls, bool)
            and issubclass(parent, int)
            and not issubclass(parent, bool)
        )
        and not (
            issubclass(cls, dt.datetime)
            and issubclass(parent, dt.date)
            and not issubclass(parent, dt.datetime)
        )
    )


@dataclass(kw_only=True, slots=True)
class IsSubclassGenError(Exception):
    cls: Any

    @override
    def __str__(self) -> str:
        return f"Argument must be a class; got {self.cls!r} of type {type(self.cls)!r}"


##


def is_tuple_type(obj: Any, /) -> bool:
    """Check if an object is a tuple type annotation."""
    return _is_annotation_of_type(obj, tuple)


##


def is_union_type(obj: Any, /) -> bool:
    """Check if an object is a union type annotation."""
    is_old_union = _is_annotation_of_type(obj, Union)  # pyright: ignore[reportDeprecated]
    return is_old_union or _is_annotation_of_type(obj, UnionType)


##


def _is_annotation_of_type(obj: Any, origin: Any, /) -> bool:
    """Check if an object is an annotation with a given origin."""
    return (get_origin(obj) is origin) or (
        isinstance(obj, TypeAliasType) and _is_annotation_of_type(obj.__value__, origin)
    )


__all__ = [
    "GetTypeClassesError",
    "GetUnionTypeClassesError",
    "IsInstanceGenError",
    "IsSubclassGenError",
    "get_literal_elements",
    "get_type_classes",
    "get_type_hints",
    "get_union_type_classes",
    "is_dict_type",
    "is_frozenset_type",
    "is_instance_gen",
    "is_list_type",
    "is_literal_type",
    "is_mapping_type",
    "is_namedtuple_class",
    "is_namedtuple_instance",
    "is_optional_type",
    "is_sequence_type",
    "is_set_type",
    "is_subclass_gen",
    "is_tuple_type",
    "is_union_type",
]



from __future__ import annotations

from collections.abc import Iterable
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path
from types import NoneType
from typing import Final, Literal

from hypothesis import given
from hypothesis.strategies import (
    booleans,
    dictionaries,
    floats,
    frozensets,
    integers,
    ip_addresses,
    lists,
    none,
    sampled_from,
    sets,
)
from pytest import raises
from whenever import (
    Date,
    DateDelta,
    DateTimeDelta,
    MonthDay,
    PlainDateTime,
    Time,
    TimeDelta,
    YearMonth,
    ZonedDateTime,
)

from tests.test_objects.objects import TruthEnum
from tests.test_typing_funcs.with_future import (
    DataClassFutureInt,
    DataClassFutureIntEven,
    DataClassFutureIntEvenOrOddTypeUnion,
    DataClassFutureIntEvenOrOddUnion,
    DataClassFutureIntOdd,
    TrueOrFalseFutureLit,
    TrueOrFalseFutureTypeLit,
)
from utilities.errors import ImpossibleCaseError
from utilities.functions import ensure_path
from utilities.hypothesis import (
    dates,
    int64s,
    month_days,
    numbers,
    paths,
    plain_datetimes,
    text_ascii,
    time_deltas,
    times,
    versions,
    year_months,
    zoned_datetimes,
)
from utilities.math import is_equal
from utilities.parse import (
    _ParseObjectExtraNonUniqueError,
    _ParseObjectParseError,
    _SerializeObjectExtraNonUniqueError,
    _SerializeObjectSerializeError,
    parse_object,
    serialize_object,
)
from utilities.sentinel import Sentinel, sentinel
from utilities.text import parse_bool
from utilities.types import Number
from utilities.version import Version


class TestSerializeAndParseObject:
    @given(bool_=booleans())
    def test_bool(self, *, bool_: bool) -> None:
        serialized = serialize_object(bool_)
        result = parse_object(bool, serialized)
        assert result is bool_

    @given(date=dates())
    def test_date(self, *, date: Date) -> None:
        serialized = serialize_object(date)
        result = parse_object(Date, serialized)
        assert result == date

    @given(mapping=dictionaries(int64s(), int64s()))
    def test_dict(self, *, mapping: dict[int, int]) -> None:
        serialized = serialize_object(mapping)
        result = parse_object(dict[int, int], serialized)
        assert result == mapping

    @given(truth=sampled_from(TruthEnum))
    def test_enum(self, *, truth: TruthEnum) -> None:
        serialized = serialize_object(truth)
        result = parse_object(TruthEnum, serialized)
        assert result is truth

    @given(float_=floats())
    def test_float(self, *, float_: float) -> None:
        serialized = serialize_object(float_)
        result = parse_object(float, serialized)
        assert is_equal(result, float_)

    @given(ints=frozensets(int64s()))
    def test_frozenset(self, *, ints: frozenset[int]) -> None:
        serialized = serialize_object(ints)
        result = parse_object(frozenset[int], serialized)
        assert result == ints

    @given(int_=integers())
    def test_int(self, *, int_: int) -> None:
        serialized = serialize_object(int_)
        result = parse_object(int, serialized)
        assert result == int_

    @given(address=ip_addresses(v=4))
    def test_ipv4_address(self, *, address: IPv4Address) -> None:
        serialized = serialize_object(address)
        result = parse_object(IPv4Address, serialized)
        assert result == address

    @given(address=ip_addresses(v=6))
    def test_ipv6_address(self, *, address: IPv6Address) -> None:
        serialized = serialize_object(address)
        result = parse_object(IPv6Address, serialized)
        assert result == address

    @given(ints=lists(int64s()))
    def test_list(self, *, ints: list[int]) -> None:
        serialized = serialize_object(ints)
        result = parse_object(list[int], serialized)
        assert result == ints

    @given(bool_=booleans())
    def test_literal_extra(self, *, bool_: bool) -> None:
        text = serialize_object(bool_)
        result = parse_object(bool, text, extra={Literal["lit"]: parse_bool})
        assert result is bool_

    @given(truth=sampled_from(["true", "false"]))
    def test_literal(self, *, truth: Literal["true", "false"]) -> None:
        result = parse_object(TrueOrFalseFutureLit, truth)
        assert result == truth

    @given(month_day=month_days())
    def test_month_day(self, *, month_day: MonthDay) -> None:
        serialized = serialize_object(month_day)
        result = parse_object(MonthDay, serialized)
        assert result == month_day

    def test_none(self) -> None:
        serialized = serialize_object(None)
        result = parse_object(None, serialized)
        assert result is None

    def test_none_type(self) -> None:
        serialized = serialize_object(None)
        result = parse_object(NoneType, serialized)
        assert result is None

    @given(number=numbers())
    def test_number(self, *, number: Number) -> None:
        serialized = serialize_object(number)
        result = parse_object(Number, serialized)
        assert result == number

    @given(path=paths())
    def test_path(self, *, path: Path) -> None:
        serialized = serialize_object(path)
        result = parse_object(Path, serialized)
        assert result == path

    @given(path=paths())
    def test_path_expanded(self, *, path: Path) -> None:
        path_use = Path("~", path)
        serialized = serialize_object(path_use)
        result = ensure_path(parse_object(Path, serialized))
        assert result == result.expanduser()

    @given(datetime=plain_datetimes())
    def test_plain_datetime(self, *, datetime: PlainDateTime) -> None:
        serialized = serialize_object(datetime)
        result = parse_object(PlainDateTime, serialized)
        assert result == datetime

    def test_nullable_number_none(self) -> None:
        serialized = serialize_object(None)
        result = parse_object(Number | None, serialized)
        assert result is None

    @given(number=numbers())
    def test_nullable_number_number(self, *, number: Number) -> None:
        serialized = serialize_object(number)
        result = parse_object(Number | None, serialized)
        assert result == number

    def test_nullable_int_none(self) -> None:
        serialized = serialize_object(None)
        result = parse_object(int | None, serialized)
        assert result is None

    @given(int_=integers())
    def test_nullable_int_int(self, *, int_: int) -> None:
        serialized = serialize_object(int_)
        result = parse_object(int | None, serialized)
        assert result == int_

    def test_sentinel(self) -> None:
        serialized = serialize_object(sentinel)
        result = parse_object(Sentinel, serialized)
        assert result is sentinel

    @given(ints=sets(int64s()))
    def test_set(self, *, ints: set[int]) -> None:
        serialized = serialize_object(ints)
        result = parse_object(set[int], serialized)
        assert result == ints

    @given(serialized=text_ascii())
    def test_to_serialized(self, *, serialized: str) -> None:
        result = parse_object(str, serialized)
        assert result == serialized

    @given(time=times())
    def test_time(self, *, time: Time) -> None:
        serialized = serialize_object(time)
        result = parse_object(Time, serialized)
        assert result == time

    @given(time_delta=time_deltas())
    def test_time_delta(self, *, time_delta: TimeDelta) -> None:
        serialized = serialize_object(time_delta)
        result = parse_object(TimeDelta, serialized)
        assert result == time_delta

    @given(x=integers(), y=integers())
    def test_tuple(self, *, x: int, y: int) -> None:
        serialized = serialize_object((x, y))
        result = parse_object(tuple[int, int], serialized)
        assert result == (x, y)

    @given(int_=integers())
    def test_type_extra(self, *, int_: int) -> None:
        serialized = serialize_object(int_)
        result = parse_object(
            DataClassFutureInt,
            serialized,
            extra={DataClassFutureInt: lambda text: DataClassFutureInt(int_=int(text))},
        )
        expected = DataClassFutureInt(int_=int_)
        assert result == expected

    @given(truth=sampled_from(["true", "false"]))
    def test_type_literal(self, *, truth: Literal["true", "false"]) -> None:
        result = parse_object(TrueOrFalseFutureTypeLit, truth)
        assert result == truth

    @given(int_=integers())
    def test_type_union_with_extra(self, *, int_: int) -> None:
        def parser(text: str, /) -> DataClassFutureIntEvenOrOddTypeUnion:
            int_ = int(text)
            match int_ % 2:
                case 0:
                    return DataClassFutureIntEven(even_int=int_)
                case 1:
                    return DataClassFutureIntOdd(odd_int=int_)
                case _:
                    raise ImpossibleCaseError(case=[f"{int_=}"])

        serialized = serialize_object(int_)
        result = parse_object(
            DataClassFutureIntEvenOrOddTypeUnion,
            serialized,
            extra={DataClassFutureIntEvenOrOddTypeUnion: parser},
        )
        match int_ % 2:
            case 0:
                expected = DataClassFutureIntEven(even_int=int_)
            case 1:
                expected = DataClassFutureIntOdd(odd_int=int_)
            case _:
                raise ImpossibleCaseError(case=[f"{int_=}"])
        assert result == expected

    @given(int_=integers())
    def test_union_with_extra(self, *, int_: int) -> None:
        def parser(text: str, /) -> DataClassFutureIntEvenOrOddUnion:
            int_ = int(text)
            match int_ % 2:
                case 0:
                    return DataClassFutureIntEven(even_int=int_)
                case 1:
                    return DataClassFutureIntOdd(odd_int=int_)
                case _:
                    raise ImpossibleCaseError(case=[f"{int_=}"])

        serialized = serialize_object(int_)
        result = parse_object(
            DataClassFutureIntEvenOrOddUnion,
            serialized,
            extra={DataClassFutureIntEvenOrOddUnion: parser},
        )
        match int_ % 2:
            case 0:
                expected = DataClassFutureIntEven(even_int=int_)
            case 1:
                expected = DataClassFutureIntOdd(odd_int=int_)
            case _:
                raise ImpossibleCaseError(case=[f"{int_=}"])
        assert result == expected

    @given(version=versions())
    def test_version(self, *, version: Version) -> None:
        serialized = serialize_object(version)
        result = parse_object(Version, serialized)
        assert result == version

    @given(year_month=year_months())
    def test_year_month(self, *, year_month: YearMonth) -> None:
        serialized = serialize_object(year_month)
        result = parse_object(YearMonth, serialized)
        assert result == year_month

    @given(datetime=zoned_datetimes())
    def test_zoned_datetime(self, *, datetime: ZonedDateTime) -> None:
        serialized = serialize_object(datetime)
        result = parse_object(ZonedDateTime, serialized)
        assert result == datetime


class TestParseObject:
    @given(text=sampled_from(["F_a_l_s_e", "T_r_u_e"]))
    def test_extra(self, *, text: str) -> None:
        def parser(text: str, /) -> bool:
            match text:
                case "F_a_l_s_e":
                    return False
                case "T_r_u_e":
                    return True
                case _:
                    raise ImpossibleCaseError(case=[f"{text=}"])

        bool_ = parse_object(bool, text, extra={bool: parser})
        match text:
            case "F_a_l_s_e":
                expected = False
            case "T_r_u_e":
                expected = True
            case _:
                raise ImpossibleCaseError(case=[f"{text=}"])
        assert bool_ is expected

    @given(text=sampled_from(["F_a_l_s_e", "T_r_u_e"]))
    def test_extra_both_exact_match_and_non_unique_parents(self, *, text: str) -> None:
        def parser(text: str, /) -> bool:
            match text:
                case "F_a_l_s_e":
                    return False
                case "T_r_u_e":
                    return True
                case _:
                    raise ImpossibleCaseError(case=[f"{text=}"])

        result = parse_object(
            bool, text, extra={bool: parser, bool | int: bool, bool | float: bool}
        )
        match text:
            case "F_a_l_s_e":
                expected = False
            case "T_r_u_e":
                expected = True
            case _:
                raise ImpossibleCaseError(case=[f"{text=}"])
        assert result is expected

    @given(value=text_ascii(min_size=10) | none())
    def test_optional_type_with_union_extra_not_used(
        self, *, value: str | None
    ) -> None:
        text = serialize_object(value)

        def parser(text: str, /) -> Number:
            return int(text)

        result = parse_object(str | None, text, extra={Number: parser})
        assert result == value

    def test_error_bool(self) -> None:
        with raises(
            _ParseObjectParseError,
            match="Unable to parse <class 'bool'>; got 'invalid'",
        ):
            _ = parse_object(bool, "invalid")

    def test_error_date(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'whenever\.Date'>; got 'invalid'",
        ):
            _ = parse_object(Date, "invalid")

    def test_error_date_delta(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'whenever\.DateDelta'>; got 'invalid'",
        ):
            _ = parse_object(DateDelta, "invalid")

    def test_error_date_time_delta(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'whenever\.DateTimeDelta'>; got 'invalid'",
        ):
            _ = parse_object(DateTimeDelta, "invalid")

    def test_error_dict_extract_group(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse dict\[int, int\]; got 'invalid'",
        ):
            _ = parse_object(dict[int, int], "invalid")

    def test_error_dict_internal(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse dict\[int, int\]; got '\{invalid=invalid\}'",
        ):
            _ = parse_object(dict[int, int], "{invalid=invalid}")

    def test_error_enum(self) -> None:
        with raises(
            _ParseObjectParseError,
            match="Unable to parse <enum 'TruthEnum'>; got 'invalid'",
        ):
            _ = parse_object(TruthEnum, "invalid")

    def test_error_extra_empty(self) -> None:
        with raises(
            _ParseObjectParseError,
            match="Unable to parse <class 'tests.test_typing_funcs.with_future.DataClassFutureInt'>; got 'invalid'",
        ):
            _ = parse_object(DataClassFutureInt, "invalid", extra={})

    @given(text=sampled_from(["F_a_l_s_e", "T_r_u_e"]))
    def test_error_extra_empty_bool_does_not_use_int(self, *, text: str) -> None:
        def parser(text: str, /) -> bool:
            match text:
                case "F_a_l_s_e":
                    return False
                case "T_r_u_e":
                    return True
                case _:
                    raise ImpossibleCaseError(case=[f"{text=}"])

        with raises(
            _ParseObjectParseError, match="Unable to parse <class 'bool'>; got '.*'"
        ):
            _ = parse_object(bool, text, extra={int: parser})

    @given(int_=integers())
    def test_error_extra_non_unique(self, *, int_: int) -> None:
        with raises(
            _ParseObjectExtraNonUniqueError,
            match="Unable to parse <class 'int'> since `extra` must contain exactly one parent class; got <class 'int'>, <class 'int'> and perhaps more",
        ):
            _ = parse_object(int, str(int_), extra={int | bool: int, int | float: int})

    def test_error_extra_union_type_extra(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse tests\.test_typing_funcs\.with_future\.DataClassFutureIntEven \| tests\.test_typing_funcs\.with_future\.DataClassFutureIntOdd; got 'invalid'",
        ):
            _ = parse_object(DataClassFutureIntEvenOrOddUnion, "invalid", extra={})

    def test_error_float(self) -> None:
        with raises(
            _ParseObjectParseError,
            match="Unable to parse <class 'float'>; got 'invalid'",
        ):
            _ = parse_object(float, "invalid")

    def test_error_frozenset_extract_group(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse frozenset\[int\]; got 'invalid'",
        ):
            _ = parse_object(frozenset[int], "invalid")

    def test_error_frozenset_internal(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse frozenset\[int\]; got '\{invalid\}'",
        ):
            _ = parse_object(frozenset[int], "{invalid}")

    def test_error_int(self) -> None:
        with raises(
            _ParseObjectParseError, match="Unable to parse <class 'int'>; got 'invalid'"
        ):
            _ = parse_object(int, "invalid")

    def test_error_list_extract_group(self) -> None:
        with raises(
            _ParseObjectParseError, match=r"Unable to parse list\[int\]; got 'invalid'"
        ):
            _ = parse_object(list[int], "invalid")

    def test_error_list_internal(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse list\[int\]; got '\[invalid\]'",
        ):
            _ = parse_object(list[int], "[invalid]")

    def test_error_month_day(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'whenever\.MonthDay'>; got 'invalid'",
        ):
            _ = parse_object(MonthDay, "invalid")

    def test_error_none(self) -> None:
        with raises(
            _ParseObjectParseError, match="Unable to parse None; got 'invalid'"
        ):
            _ = parse_object(None, "invalid")

    def test_error_none_type(self) -> None:
        with raises(
            _ParseObjectParseError,
            match="Unable to parse <class 'NoneType'>; got 'invalid'",
        ):
            _ = parse_object(NoneType, "invalid")

    def test_error_nullable_int(self) -> None:
        with raises(
            _ParseObjectParseError, match=r"Unable to parse int \| None; got 'invalid'"
        ):
            _ = parse_object(int | None, "invalid")

    def test_error_nullable_not_type(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse collections\.abc\.Iterable\[None\] \| None; got 'invalid'",
        ):
            _ = parse_object(Iterable[None] | None, "invalid")

    def test_error_number(self) -> None:
        with raises(
            _ParseObjectParseError, match=r"Unable to parse Number; got 'invalid'"
        ):
            _ = parse_object(Number, "invalid")

    def test_error_plain_datetime(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'whenever\.PlainDateTime'>; got 'invalid'",
        ):
            _ = parse_object(PlainDateTime, "invalid")

    def test_error_sentinel(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'utilities\.sentinel\.Sentinel'>; got 'invalid'",
        ):
            _ = parse_object(Sentinel, "invalid")

    def test_error_set_extract_group(self) -> None:
        with raises(
            _ParseObjectParseError, match=r"Unable to parse set\[int\]; got 'invalid'"
        ):
            _ = parse_object(set[int], "invalid")

    def test_error_set_internal(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse set\[int\]; got '\{invalid\}'",
        ):
            _ = parse_object(set[int], "{invalid}")

    def test_error_time(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'whenever\.Time'>; got 'invalid'",
        ):
            _ = parse_object(Time, "invalid")

    def test_error_time_delta(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'whenever\.TimeDelta'>; got 'invalid'",
        ):
            _ = parse_object(TimeDelta, "invalid")

    def test_error_tuple_extract_group(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse tuple\[int, int\]; got 'invalid'",
        ):
            _ = parse_object(tuple[int, int], "invalid")

    def test_error_tuple_internal(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse tuple\[int, int\]; got '\(invalid,invalid\)'",
        ):
            _ = parse_object(tuple[int, int], "(invalid,invalid)")

    def test_error_tuple_inconsistent_args_and_serializeds(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse tuple\[int, int\]; got '\(serialized1, serialized2, serialized3\)'",
        ):
            _ = parse_object(tuple[int, int], "(serialized1, serialized2, serialized3)")

    def test_error_type_not_implemented(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'tests\.test_typing_funcs\.with_future\.DataClassFutureInt'>; got 'invalid'",
        ):
            _ = parse_object(DataClassFutureInt, "invalid")

    def test_error_union_not_implemented(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse tests\.test_typing_funcs\.with_future\.DataClassFutureIntEven \| tests\.test_typing_funcs\.with_future\.DataClassFutureIntOdd; got 'invalid'",
        ):
            _ = parse_object(DataClassFutureIntEvenOrOddUnion, "invalid")

    def test_error_version(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'utilities\.version\.Version'>; got 'invalid'",
        ):
            _ = parse_object(Version, "invalid")

    def test_error_year_month(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'whenever\.YearMonth'>; got 'invalid'",
        ):
            _ = parse_object(YearMonth, "invalid")

    def test_error_zoned_datetime(self) -> None:
        with raises(
            _ParseObjectParseError,
            match=r"Unable to parse <class 'whenever\.ZonedDateTime'>; got 'invalid'",
        ):
            _ = parse_object(ZonedDateTime, "invalid")


class TestSerializeObject:
    @given(bool_=booleans())
    def test_bool_custom(self, *, bool_: bool) -> None:
        def serializer(bool_: bool, /) -> str:  # noqa: FBT001
            match bool_:
                case True:
                    return "1"
                case False:
                    return "0"

        serialized = serialize_object(bool_, extra={bool: serializer})
        match bool_:
            case True:
                expected = "1"
            case False:
                expected = "0"
        assert serialized == expected

    @given(bool_=booleans())
    def test_bool_extra_not_used(self, *, bool_: bool) -> None:
        def serializer(int_: int, /) -> str:
            return f"({int_})"

        serialized = serialize_object(bool_, extra={int: serializer})
        expected = str(bool_)
        assert serialized == expected

    @given(int_=integers())
    def test_type_with_extra(self, *, int_: int) -> None:
        obj = DataClassFutureInt(int_=int_)

        def serializer(obj: DataClassFutureInt, /) -> str:
            return str(obj.int_)

        serialized = serialize_object(obj, extra={DataClassFutureInt: serializer})
        expected = str(int_)
        assert serialized == expected

    @given(int_=integers())
    def test_type_union_with_extra(self, *, int_: int) -> None:
        match int_ % 2:
            case 0:
                obj = DataClassFutureIntEven(even_int=int_)
            case 1:
                obj = DataClassFutureIntOdd(odd_int=int_)
            case _:
                raise ImpossibleCaseError(case=[f"{int_=}"])

        def serializer(obj: DataClassFutureIntEvenOrOddTypeUnion, /) -> str:
            match obj:
                case DataClassFutureIntEven():
                    return str(obj.even_int)
                case DataClassFutureIntOdd():
                    return str(obj.odd_int)

        serialized = serialize_object(
            obj, extra={DataClassFutureIntEvenOrOddTypeUnion: serializer}
        )
        expected = str(int_)
        assert serialized == expected

    @given(int_=integers())
    def test_union_with_extra(self, *, int_: int) -> None:
        match int_ % 2:
            case 0:
                obj = DataClassFutureIntEven(even_int=int_)
            case 1:
                obj = DataClassFutureIntOdd(odd_int=int_)
            case _:
                raise ImpossibleCaseError(case=[f"{int_=}"])

        def serializer(obj: DataClassFutureIntEvenOrOddUnion, /) -> str:
            match obj:
                case DataClassFutureIntEven():
                    return str(obj.even_int)
                case DataClassFutureIntOdd():
                    return str(obj.odd_int)

        serialized = serialize_object(
            obj, extra={DataClassFutureIntEvenOrOddUnion: serializer}
        )
        expected = str(int_)
        assert serialized == expected

    def test_error_extra_empty(self) -> None:
        with raises(
            _SerializeObjectSerializeError,
            match=r"Unable to serialize object typing\.Final of type <class 'typing\._SpecialForm'>",
        ):
            _ = serialize_object(Final, extra={})

    @given(bool_=booleans())
    def test_error_extra_non_unique(self, *, bool_: bool) -> None:
        with raises(
            _SerializeObjectExtraNonUniqueError,
            match=r"Unable to serialize object (True|False) of type <class 'bool'> since `extra` must contain exactly one parent class; got <class 'str'>, <class 'str'> and perhaps more",
        ):
            _ = serialize_object(bool_, extra={bool | int: str, bool | float: str})

    def test_error_not_implemented(self) -> None:
        with raises(_SerializeObjectSerializeError):
            _ = serialize_object(Final)



from __future__ import annotations

from ipaddress import IPv4Address

from dns.resolver import resolve

from utilities.functools import cache


@cache
def nslookup(address: str, /) -> list[IPv4Address]:
    """Look up a set of addresses."""
    ans = resolve(address)  # skipif-ci-and-mac
    return [IPv4Address(str(rd)) for rd in ans]  # skipif-ci-and-mac


__all__ = ["nslookup"]



from __future__ import annotations

from collections.abc import Collection, Iterator
from re import escape
from typing import TYPE_CHECKING, Any, ClassVar, TypeGuard

from pytest import mark, param, raises

from utilities.more_itertools import (
    BucketMappingError,
    Split,
    bucket_mapping,
    partition_list,
    partition_typeguard,
    peekable,
    yield_splits,
)
from utilities.text import strip_and_dedent

if TYPE_CHECKING:
    from collections.abc import Iterable


class TestBucketMapping:
    iterable: ClassVar[list[str]] = ["a1", "b1", "c1", "a2", "b2", "c2", "b3"]
    iterable_unique: ClassVar[list[str]] = ["a1", "b2", "c3"]

    def test_main(self) -> None:
        mapping = bucket_mapping(self.iterable, lambda x: x[0])
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, Iterator)
            assert not isinstance(value, Collection)
        assert list(mapping["a"]) == ["a1", "a2"]
        assert list(mapping["b"]) == ["b1", "b2", "b3"]
        assert list(mapping["c"]) == ["c1", "c2"]

    def test_list(self) -> None:
        mapping = bucket_mapping(self.iterable, lambda x: x[0], post="list")
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, list)
        assert mapping["a"] == ["a1", "a2"]
        assert mapping["b"] == ["b1", "b2", "b3"]
        assert mapping["c"] == ["c1", "c2"]

    def test_tuple(self) -> None:
        mapping = bucket_mapping(self.iterable, lambda x: x[0], post="tuple")
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, tuple)
        assert mapping["a"] == ("a1", "a2")
        assert mapping["b"] == ("b1", "b2", "b3")
        assert mapping["c"] == ("c1", "c2")

    def test_set(self) -> None:
        mapping = bucket_mapping(self.iterable, lambda x: x[0], post="set")
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, set)
        assert mapping["a"] == {"a1", "a2"}
        assert mapping["b"] == {"b1", "b2", "b3"}
        assert mapping["c"] == {"c1", "c2"}

    def test_frozenset(self) -> None:
        mapping = bucket_mapping(self.iterable, lambda x: x[0], post="frozenset")
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, frozenset)
        assert mapping["a"] == frozenset({"a1", "a2"})
        assert mapping["b"] == frozenset({"b1", "b2", "b3"})
        assert mapping["c"] == frozenset({"c1", "c2"})

    def test_unique(self) -> None:
        mapping = bucket_mapping(self.iterable_unique, lambda x: x[0], post="unique")
        expected = {"a": "a1", "b": "b2", "c": "c3"}
        assert mapping == expected

    def test_pre(self) -> None:
        mapping = bucket_mapping(
            self.iterable, lambda x: x[0], pre=lambda x: int(x[-1])
        )
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, Iterator)
            assert not isinstance(value, Collection)
        assert list(mapping["a"]) == [1, 2]
        assert list(mapping["b"]) == [1, 2, 3]
        assert list(mapping["c"]) == [1, 2]

    def test_pre_and_list(self) -> None:
        mapping = bucket_mapping(
            self.iterable, lambda x: x[0], pre=lambda x: int(x[-1]), post="list"
        )
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, list)
        assert mapping["a"] == [1, 2]
        assert mapping["b"] == [1, 2, 3]
        assert mapping["c"] == [1, 2]

    def test_pre_and_tuple(self) -> None:
        mapping = bucket_mapping(
            self.iterable, lambda x: x[0], pre=lambda x: int(x[-1]), post="tuple"
        )
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, tuple)
        assert mapping["a"] == (1, 2)
        assert mapping["b"] == (1, 2, 3)
        assert mapping["c"] == (1, 2)

    def test_pre_and_set(self) -> None:
        mapping = bucket_mapping(
            self.iterable, lambda x: x[0], pre=lambda x: int(x[-1]), post="set"
        )
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, set)
        assert mapping["a"] == {1, 2}
        assert mapping["b"] == {1, 2, 3}
        assert mapping["c"] == {1, 2}

    def test_pre_and_frozenset(self) -> None:
        mapping = bucket_mapping(
            self.iterable, lambda x: x[0], pre=lambda x: int(x[-1]), post="frozenset"
        )
        assert set(mapping) == {"a", "b", "c"}
        for value in mapping.values():
            assert isinstance(value, frozenset)
        assert mapping["a"] == frozenset({1, 2})
        assert mapping["b"] == frozenset({1, 2, 3})
        assert mapping["c"] == frozenset({1, 2})

    def test_pre_and_unique(self) -> None:
        mapping = bucket_mapping(
            self.iterable_unique,
            lambda x: x[0],
            pre=lambda x: int(x[-1]),
            post="unique",
        )
        expected = {"a": 1, "b": 2, "c": 3}
        assert mapping == expected

    def test_error_unique(self) -> None:
        with raises(
            BucketMappingError,
            match=escape(
                "Buckets must contain exactly one item each; got 'a' (#1: 'a1', #2: 'a2'), 'b' (#1: 'b1', #2: 'b2'), 'c' (#1: 'c1', #2: 'c2')"
            ),
        ):
            _ = bucket_mapping(self.iterable, lambda x: x[0], post="unique")

    def test_error_pre_and_unique(self) -> None:
        with raises(
            BucketMappingError,
            match=escape(
                "Buckets must contain exactly one item each; got 'a' (#1: 1, #2: 2), 'b' (#1: 1, #2: 2), 'c' (#1: 1, #2: 2)"
            ),
        ):
            _ = bucket_mapping(
                self.iterable, lambda x: x[0], pre=lambda x: int(x[-1]), post="unique"
            )


class TestPartitionList:
    def test_main(self) -> None:
        def is_even(x: int, /) -> bool:
            return x % 2 == 0

        iterable = [1, 2, 3, 4]
        false, true = partition_list(is_even, iterable)
        assert isinstance(false, list)
        assert false == [1, 3]
        assert isinstance(true, list)
        assert true == [2, 4]


class TestPartitionTypeguard:
    def test_main(self) -> None:
        def is_int(x: Any, /) -> TypeGuard[int]:
            return isinstance(x, int)

        iterable = [1, 2.0, 3, 4.0]
        false, true = partition_typeguard(is_int, iterable)
        for el in false:
            assert isinstance(el, int | float)
        for el in true:
            assert isinstance(el, int)


class TestPeekable:
    def test_dropwhile(self) -> None:
        it = peekable(range(10))
        it.dropwhile(lambda x: x <= 4)
        assert it.peek() == 5
        result = list(it)
        expected = [5, 6, 7, 8, 9]
        assert result == expected

    def test_iter(self) -> None:
        it = peekable(range(10))
        values: list[int] = []
        for value in it:
            assert isinstance(value, int)
            values.append(value)
        assert len(values) == 10

    def test_next(self) -> None:
        it = peekable(range(10))
        value = next(it)
        assert isinstance(value, int)

    def test_peek_non_empty(self) -> None:
        it = peekable(range(10))
        value = it.peek()
        assert isinstance(value, int)

    def test_peek_empty_without_default(self) -> None:
        it: peekable[int] = peekable([])
        with raises(StopIteration):
            _ = it.peek()

    def test_peek_empty_with_default(self) -> None:
        it: peekable[int] = peekable([])
        value = it.peek(default="default")
        assert isinstance(value, str)

    def test_takewhile(self) -> None:
        it = peekable(range(10))
        result1 = list(it.takewhile(lambda x: x <= 4))
        expected1 = [0, 1, 2, 3, 4]
        assert result1 == expected1
        assert it.peek() == 5
        result2 = list(it)
        expected2 = [5, 6, 7, 8, 9]
        assert result2 == expected2

    def test_combined(self) -> None:
        it = peekable(range(10))
        result1 = list(it.takewhile(lambda x: x <= 2))
        expected1 = [0, 1, 2]
        assert result1 == expected1
        assert it.peek() == 3
        it.dropwhile(lambda x: x <= 4)
        assert it.peek() == 5
        result2 = list(it.takewhile(lambda x: x <= 6))
        expected2 = [5, 6]
        assert result2 == expected2
        result3 = list(it)
        expected3 = [7, 8, 9]
        assert result3 == expected3


class TestYieldSplits:
    @mark.parametrize(
        ("iterable", "head", "tail", "min_frac", "freq", "expected"),
        [
            param(
                "abcde",
                3,
                1,
                None,
                None,
                [
                    Split(head=["a", "b", "c"], tail=["d"]),
                    Split(head=["b", "c", "d"], tail=["e"]),
                ],
                id="3/1",
            ),
            param(
                "abcde",
                3,
                1,
                0.4,
                None,
                [
                    Split(head=["a", "b"], tail=["c"]),
                    Split(head=["a", "b", "c"], tail=["d"]),
                    Split(head=["b", "c", "d"], tail=["e"]),
                ],
                id="3/1, min-frac=0.4",
            ),
            param(
                "abcdefg",
                3,
                2,
                None,
                None,
                [
                    Split(head=["a", "b", "c"], tail=["d", "e"]),
                    Split(head=["c", "d", "e"], tail=["f", "g"]),
                ],
                id="3/2, clean tail",
            ),
            param(
                "abcdefgh",
                3,
                2,
                None,
                None,
                [
                    Split(head=["a", "b", "c"], tail=["d", "e"]),
                    Split(head=["c", "d", "e"], tail=["f", "g"]),
                    Split(head=["e", "f", "g"], tail=["h"]),
                ],
                id="3/2, truncated tail",
            ),
            param(
                "abcdefgh",
                3,
                2,
                None,
                1,
                [
                    Split(head=["a", "b", "c"], tail=["d", "e"]),
                    Split(head=["b", "c", "d"], tail=["e", "f"]),
                    Split(head=["c", "d", "e"], tail=["f", "g"]),
                    Split(head=["d", "e", "f"], tail=["g", "h"]),
                    Split(head=["e", "f", "g"], tail=["h"]),
                ],
                id="3/2, freq=1",
            ),
            param("abc", 5, 1, None, None, [], id="len(iterable) < head"),
            param("abc", 1, 5, None, None, [], id="len(iterable) < tail"),
        ],
    )
    def test_main(
        self,
        *,
        iterable: Iterable[str],
        head: int,
        tail: int,
        min_frac: float | None,
        freq: int | None,
        expected: list[Split[list[str]]],
    ) -> None:
        splits = list(yield_splits(iterable, head, tail, min_frac=min_frac, freq=freq))
        assert splits == expected

    def test_repr(self) -> None:
        split = Split(head=["a", "b", "c"], tail=["d"])
        result = repr(split)
        expected = strip_and_dedent(
            """
            Split(
                head=
                    ['a', 'b', 'c']
                tail=
                    ['d']
            )
            """
        )
        assert result == expected



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING, Literal, assert_never, override

from sqlalchemy import Table
from sqlalchemy.orm import DeclarativeBase

from utilities.asyncio import stream_command
from utilities.iterables import always_iterable
from utilities.logging import get_logger
from utilities.os import temp_environ
from utilities.pathlib import ensure_suffix
from utilities.sqlalchemy import get_table_name
from utilities.timer import Timer
from utilities.types import PathLike

if TYPE_CHECKING:
    from sqlalchemy import URL

    from utilities.sqlalchemy import TableOrORMInstOrClass
    from utilities.types import LoggerOrName, MaybeListStr, MaybeSequence, PathLike


type _PGDumpFormat = Literal["plain", "custom", "directory", "tar"]


async def pg_dump(
    url: URL,
    path: PathLike,
    /,
    *,
    format_: _PGDumpFormat = "plain",
    jobs: int | None = None,
    data_only: bool = False,
    clean: bool = False,
    schema: MaybeListStr | None = None,
    schema_exc: MaybeListStr | None = None,
    table: MaybeSequence[TableOrORMInstOrClass | str] | None = None,
    table_exc: MaybeSequence[TableOrORMInstOrClass | str] | None = None,
    inserts: bool = False,
    on_conflict_do_nothing: bool = False,
    docker: str | None = None,
    dry_run: bool = False,
    logger: LoggerOrName | None = None,
) -> None:
    """Run `pg_dump`."""
    path = _path_pg_dump(path, format_=format_)
    path.parent.mkdir(parents=True, exist_ok=True)
    cmd = _build_pg_dump(
        url,
        path,
        format_=format_,
        jobs=jobs,
        data_only=data_only,
        clean=clean,
        schema=schema,
        schema_exc=schema_exc,
        table=table,
        table_exc=table_exc,
        inserts=inserts,
        on_conflict_do_nothing=on_conflict_do_nothing,
        docker=docker,
    )
    if dry_run:
        if logger is not None:
            get_logger(logger=logger).info("Would run %r", str(cmd))
        return
    with temp_environ(PGPASSWORD=url.password), Timer() as timer:  # pragma: no cover
        try:
            output = await stream_command(cmd)
        except KeyboardInterrupt:
            if logger is not None:
                get_logger(logger=logger).info(
                    "Cancelled backup to %r after %s", str(path), timer
                )
            rmtree(path, ignore_errors=True)
        else:
            match output.return_code:
                case 0:
                    if logger is not None:
                        get_logger(logger=logger).info(
                            "Backup to %r finished after %s", str(path), timer
                        )
                case _:
                    if logger is not None:
                        get_logger(logger=logger).exception(
                            "Backup to %r failed after %s\nstderr:\n%s",
                            str(path),
                            timer,
                            output.stderr,
                        )
                    rmtree(path, ignore_errors=True)


def _build_pg_dump(
    url: URL,
    path: PathLike,
    /,
    *,
    format_: _PGDumpFormat = "plain",
    jobs: int | None = None,
    data_only: bool = False,
    clean: bool = False,
    schema: MaybeListStr | None = None,
    schema_exc: MaybeListStr | None = None,
    table: MaybeSequence[TableOrORMInstOrClass | str] | None = None,
    table_exc: MaybeSequence[TableOrORMInstOrClass | str] | None = None,
    inserts: bool = False,
    on_conflict_do_nothing: bool = False,
    docker: str | None = None,
) -> str:
    database, host, port = _extract_url(url)
    path = _path_pg_dump(path, format_=format_)
    parts: list[str] = [
        "pg_dump",
        # general options
        f"--dbname={database}",
        f"--file={str(path)!r}",
        f"--format={format_}",
        "--verbose",
        # output options
        "--large-objects",
        "--create",
        "--no-owner",
        "--no-privileges",
        # connection options
        f"--host={host}",
        f"--port={port}",
        "--no-password",
    ]
    if (format_ == "directory") and (jobs is not None):
        parts.append(f"--jobs={jobs}")
    if data_only:
        parts.append("--data-only")
    if clean:
        parts.extend(["--clean", "--if-exists"])
    if schema is not None:
        parts.extend([f"--schema={s}" for s in always_iterable(schema)])
    if schema_exc is not None:
        parts.extend([f"--exclude-schema={s}" for s in always_iterable(schema_exc)])
    if table is not None:
        parts.extend([f"--table={_get_table_name(t)}" for t in always_iterable(table)])
    if table_exc is not None:
        parts.extend([
            f"--exclude-table={_get_table_name(t)}" for t in always_iterable(table_exc)
        ])
    if inserts:
        parts.append("--inserts")
    if on_conflict_do_nothing:
        parts.append("--on-conflict-do-nothing")
    if url.username is not None:
        parts.append(f"--username={url.username}")
    if docker is not None:
        parts = _wrap_docker(parts, docker)
    return " ".join(parts)


def _path_pg_dump(path: PathLike, /, *, format_: _PGDumpFormat = "plain") -> Path:
    match format_:
        case "plain":
            suffix = ".sql"
        case "custom":
            suffix = ".pgdump"
        case "directory":
            suffix = None
        case "tar":
            suffix = ".tar"
        case _ as never:
            assert_never(never)
    path = Path(path)
    if suffix is not None:
        path = ensure_suffix(path, suffix)
    return path


##


async def restore(
    url: URL,
    path: PathLike,
    /,
    *,
    psql: bool = False,
    database: str | None = None,
    data_only: bool = False,
    clean: bool = False,
    jobs: int | None = None,
    schema: MaybeListStr | None = None,
    schema_exc: MaybeListStr | None = None,
    table: MaybeSequence[TableOrORMInstOrClass | str] | None = None,
    docker: str | None = None,
    dry_run: bool = False,
    logger: LoggerOrName | None = None,
) -> None:
    """Run `pg_restore`/`psql`."""
    cmd = _build_pg_restore_or_psql(
        url,
        path,
        psql=psql,
        database=database,
        data_only=data_only,
        clean=clean,
        jobs=jobs,
        schema=schema,
        schema_exc=schema_exc,
        table=table,
        docker=docker,
    )
    if dry_run:
        if logger is not None:
            get_logger(logger=logger).info("Would run %r", str(cmd))
        return
    with temp_environ(PGPASSWORD=url.password), Timer() as timer:  # pragma: no cover
        try:
            output = await stream_command(cmd)
        except KeyboardInterrupt:
            if logger is not None:
                get_logger(logger=logger).info(
                    "Cancelled restore from %r after %s", str(path), timer
                )
        else:
            match output.return_code:
                case 0:
                    if logger is not None:
                        get_logger(logger=logger).info(
                            "Restore from %r finished after %s", str(path), timer
                        )
                case _:
                    if logger is not None:
                        get_logger(logger=logger).exception(
                            "Restore from %r failed after %s\nstderr:\n%s",
                            str(path),
                            timer,
                            output.stderr,
                        )


##


def _build_pg_restore_or_psql(
    url: URL,
    path: PathLike,
    /,
    *,
    psql: bool = False,
    database: str | None = None,
    data_only: bool = False,
    clean: bool = False,
    jobs: int | None = None,
    schema: MaybeListStr | None = None,
    schema_exc: MaybeListStr | None = None,
    table: MaybeSequence[TableOrORMInstOrClass | str] | None = None,
    docker: str | None = None,
) -> str:
    path = Path(path)
    if (path.suffix == ".sql") or psql:
        return _build_psql(url, path, database=database, docker=docker)
    return _build_pg_restore(
        url,
        path,
        database=database,
        data_only=data_only,
        clean=clean,
        jobs=jobs,
        schemas=schema,
        schemas_exc=schema_exc,
        tables=table,
        docker=docker,
    )


def _build_pg_restore(
    url: URL,
    path: PathLike,
    /,
    *,
    database: str | None = None,
    data_only: bool = False,
    clean: bool = False,
    jobs: int | None = None,
    schemas: MaybeListStr | None = None,
    schemas_exc: MaybeListStr | None = None,
    tables: MaybeSequence[TableOrORMInstOrClass | str] | None = None,
    docker: str | None = None,
) -> str:
    """Run `pg_restore`."""
    url_database, host, port = _extract_url(url)
    database_use = url_database if database is None else database
    parts: list[str] = [
        "pg_restore",
        # general options
        f"--dbname={database_use}",
        "--verbose",
        # restore options
        "--exit-on-error",
        "--no-owner",
        "--no-privileges",
        # connection options
        f"--host={host}",
        f"--port={port}",
        "--no-password",
    ]
    if data_only:
        parts.append("--data-only")
    if clean:
        parts.extend(["--clean", "--if-exists"])
    if jobs is not None:
        parts.append(f"--jobs={jobs}")
    if schemas is not None:
        parts.extend([f"--schema={s}" for s in always_iterable(schemas)])
    if schemas_exc is not None:
        parts.extend([f"--exclude-schema={s}" for s in always_iterable(schemas_exc)])
    if tables is not None:
        parts.extend([f"--table={_get_table_name(t)}" for t in always_iterable(tables)])
    if url.username is not None:
        parts.append(f"--username={url.username}")
    if docker is not None:
        parts = _wrap_docker(parts, docker)
    parts.append(str(path))
    return " ".join(parts)


def _build_psql(
    url: URL,
    path: PathLike,
    /,
    *,
    database: str | None = None,
    docker: str | None = None,
) -> str:
    """Run `psql`."""
    url_database, host, port = _extract_url(url)
    database_use = url_database if database is None else database
    parts: list[str] = [
        "psql",
        # general options
        f"--dbname={database_use}",
        f"--file={str(path)!r}",
        # connection options
        f"--host={host}",
        f"--port={port}",
        "--no-password",
    ]
    if url.username is not None:
        parts.append(f"--username={url.username}")
    if docker is not None:
        parts = _wrap_docker(parts, docker)
    return " ".join(parts)


##


def _extract_url(url: URL, /) -> tuple[str, str, int]:
    if url.database is None:
        raise _ExtractURLDatabaseError(url=url)
    if url.host is None:
        raise _ExtractURLHostError(url=url)
    if url.port is None:
        raise _ExtractURLPortError(url=url)
    return url.database, url.host, url.port


@dataclass(kw_only=True, slots=True)
class ExtractURLError(Exception):
    url: URL


@dataclass(kw_only=True, slots=True)
class _ExtractURLDatabaseError(ExtractURLError):
    @override
    def __str__(self) -> str:
        return f"Expected URL to contain a 'database'; got {self.url}"


@dataclass(kw_only=True, slots=True)
class _ExtractURLHostError(ExtractURLError):
    @override
    def __str__(self) -> str:
        return f"Expected URL to contain a 'host'; got {self.url}"


@dataclass(kw_only=True, slots=True)
class _ExtractURLPortError(ExtractURLError):
    @override
    def __str__(self) -> str:
        return f"Expected URL to contain a 'port'; got {self.url}"


def _get_table_name(obj: TableOrORMInstOrClass | str, /) -> str:
    match obj:
        case Table() | DeclarativeBase() | type() as table_or_orm:
            return get_table_name(table_or_orm)
        case str() as name:
            return name
        case _ as never:
            assert_never(never)


def _wrap_docker(parts: list[str], container: str, /) -> list[str]:
    return ["docker", "exec", "-it", container, *parts]


__all__ = ["ExtractURLError", "pg_dump", "restore"]



from __future__ import annotations

from asyncio import run
from logging import getLogger
from random import randint
from typing import TYPE_CHECKING

from redis.asyncio import Redis

from utilities.asyncio import sleep_td
from utilities.logging import setup_logging
from utilities.pathlib import get_repo_root
from utilities.pottery import extend_lock, try_yield_coroutine_looper
from utilities.whenever import SECOND

if TYPE_CHECKING:
    from pottery import AIORedlock

_LOGGER = getLogger(__name__)


async def script(*, lock: AIORedlock | None = None) -> None:
    total = 1000
    fail = 30
    success = 3
    while True:
        n = randint(0, total)
        if n < fail:
            _LOGGER.info("n = %d; failing...", n)
            msg = f"n = {n}; failure"
            raise ValueError(msg)
        if fail <= n < (fail + success):
            _LOGGER.info("n = %d; succeeding...", n)
            return
        _LOGGER.info("n = %d", n)
        await extend_lock(lock=lock)
        await sleep_td(SECOND / 3)


async def service() -> None:
    redis = Redis()
    async with try_yield_coroutine_looper(
        redis,
        "utilities-test",
        num=1,
        timeout_release=5 * SECOND,
        logger=_LOGGER,
        sleep_error=4 * SECOND,
    ) as looper:
        if looper is not None:
            result = await looper(script, lock=looper.lock)
            _LOGGER.info("script %s", "succeeded" if result else "failed")


def main() -> None:
    setup_logging(logger=_LOGGER, files_dir=get_repo_root().joinpath(".logs"))
    run(service())


if __name__ == "__main__":
    main()
